"""Polynomial factorization routines in characteristic zero."""

import functools
import math
import operator

from ..config import query
from ..ntheory import factorint, isprime, nextprime
from ..ntheory.modular import symmetric_residue
from ..utilities import subsets
from .polyerrors import (CoercionFailed, DomainError, EvaluationFailed,
                         ExtraneousFactors)
from .polyutils import _sort_factors


class _Factor:
    """Mixin class for factorization routines."""

    def _trial_division(self, f, factors):
        result = []

        for factor in factors:
            k = 0

            while f:
                q, r = divmod(f, factor)

                if not r:
                    f, k = q, k + 1
                else:
                    break

            result.append((factor, k))

        return _sort_factors(result)

    def _aa_factor_trager(self, f):
        """Factor multivariate polynomials over algebraic number fields."""
        domain = self.domain

        lc, f = f.LC, f.monic()

        if f.is_ground:
            return lc, []

        f, F = f.sqf_part(), f
        s, g, r = f.sqf_norm()

        _, factors = r.factor_list()

        if len(factors) == 1:
            factors = [f]
        else:
            for i, (factor, _) in enumerate(factors):
                h = factor.set_domain(domain)
                h, _, g = self.cofactors(h, g)
                h = h.compose({x: x + s*domain.unit for x in self.gens})
                factors[i] = h

        return lc, self._trial_division(F, factors)

    def _zz_factor(self, f):
        """
        Factor (non square-free) polynomials in `Z[X]`.

        Given a multivariate polynomial `f` in `Z[X]` computes its complete
        factorization `f_1, ..., f_n` into irreducibles over integers::

            f = content(f) f_1**k_1 ... f_n**k_n

        The factorization is computed by reducing the input polynomial
        into a primitive square-free polynomial and factoring it using
        Zassenhaus or Enhanced Extended Zassenhaus (EEZ) algorithm. Trial
        division is used to recover the multiplicities of factors.

        The result is returned as a tuple consisting of::

            (content(f), [(f_1, k_1), ..., (f_n, k_n))

        Examples
        ========

        >>> _, x = ring('x', ZZ)

        >>> (2*x**4 - 2).factor_list()
        (2, [(x - 1, 1), (x + 1, 1), (x**2 + 1, 1)])

        >>> _, x, y = ring('x y', ZZ)

        >>> (2*x**2 - 2*y**2).factor_list()
        (2, [(x - y, 1), (x + y, 1)])

        References
        ==========

        * :cite:`Gathen1999modern`

        """
        domain = self.domain

        if self.is_univariate:
            cont, g = f.primitive()

            n = g.degree()

            if n <= 0:
                return cont, []
            elif n == 1:
                return cont, [(g, 1)]

            if query('USE_IRREDUCIBLE_IN_FACTOR'):
                if self._zz_irreducible_p(g):
                    return cont, [(g, 1)]

            g = g.sqf_part()
            H = None

            if query('USE_CYCLOTOMIC_FACTOR'):
                H = self._zz_cyclotomic_factor(g)

            if H is None:
                H = self._zz_zassenhaus(g)

            factors = self._trial_division(f, H)
            return cont, factors

        if not f:
            return domain.zero, []

        cont, g = f.primitive()

        if g.is_ground:
            return cont, []

        g = g.eject(*self.gens[1:])
        G, g = g.primitive()
        g = g.inject()

        factors = []

        if g.degree() > 0:
            g = g.sqf_part()
            H = self._zz_wang(g)
            factors = self._trial_division(f, H)

        new_ring = G.ring

        for g, k in new_ring._zz_factor(G)[1]:
            factors.insert(0, (g.set_ring(self), k))

        return cont, _sort_factors(factors)

    def _zz_mignotte_bound(self, f):
        """Mignotte bound for multivariate polynomials in `Z[X]`."""
        domain = self.domain

        a = f.max_norm()
        b = abs(f.LC)
        n = sum(f.degree(x) for x in self.gens)

        return domain.sqrt(domain(n + 1))*2**n*a*b

    def _zz_zassenhaus(self, f):
        """Factor primitive square-free polynomial in `Z[x]`."""
        domain = self.domain

        assert self.is_univariate
        assert domain.is_IntegerRing

        n = f.degree()

        if n == 1:
            return [f]

        fc = f[1]
        A = f.max_norm()
        b = f.LC
        B = int(self._zz_mignotte_bound(f))
        C = int((n + 1)**(2*n)*A**(2*n - 1))
        gamma = math.ceil(2*math.log(C, 2))
        bound = int(2*gamma*math.log(gamma))
        a = []

        # choose a prime number `p` such that `f` be square free in Z_p
        # if there are many factors in Z_p, choose among a few different `p`
        # the one with fewer factors
        for p in range(3, bound + 1):  # pragma: no branch
            if not isprime(p) or b % p == 0:
                continue

            p = domain.convert(p)
            p_domain = domain.finite_field(p)

            F = f.set_domain(p_domain)

            if not F.is_squarefree:
                continue

            F = F.monic()
            fsqfx = [_ for _, k in F.factor_list()[1]]

            a.append((p, [_.set_domain(domain) for _ in fsqfx]))
            if len(fsqfx) < 15 or len(a) > 4:
                break
        p, fsqf = min(a, key=lambda x: len(x[1]))

        l = math.ceil(math.log(2*B + 1, p))
        g = self._zz_hensel_lift(p, f, fsqf, l)

        sorted_T = range(len(g))
        T = set(sorted_T)
        factors, s = [], 1
        pl = p**l

        while 2*s <= len(T):
            for S in subsets(sorted_T, s):
                # lift the constant coefficient of the product `G` of the factors
                # in the subset `S`; if it is does not divide `fc`, `G` does
                # not divide the input polynomial

                if b == 1:
                    q = 1
                    for i in S:
                        q = q*g[i][1]
                    q = q % pl
                    qs = symmetric_residue(q, pl)
                    if qs and fc % qs != 0:
                        continue
                else:
                    G = self.ground_new(b)
                    for i in S:
                        G *= g[i]
                    G = G.trunc_ground(pl)
                    _, G = G.primitive()
                    q = G[1]
                    if q and fc % q != 0:
                        continue

                H = self.ground_new(b)
                S = set(S)
                T_S = T - S

                if b == 1:
                    G = self.ground_new(b)
                    for i in S:
                        G *= g[i]
                    G = G.trunc_ground(pl)

                for i in T_S:
                    H *= g[i]

                H = H.trunc_ground(pl)

                G_norm = G.l1_norm()
                H_norm = H.l1_norm()

                if G_norm*H_norm <= B:
                    T = T_S
                    sorted_T = [i for i in sorted_T if i not in S]

                    G = G.primitive()[1]
                    f = H.primitive()[1]

                    factors.append(G)
                    b = f.LC

                    break
            else:
                s += 1

        return factors + [f]

    def _zz_factor_sqf(self, f):
        """Factor square-free (non-primitive) polynomials in `Z[x]`."""
        domain = self.domain

        assert self.is_univariate
        assert domain.is_IntegerRing

        cont, g = f.primitive()

        n = g.degree()

        if n <= 0:
            return cont, []
        elif n == 1:
            return cont, [g]

        if query('USE_IRREDUCIBLE_IN_FACTOR'):
            if self._zz_irreducible_p(g):
                return cont, [g]

        factors = None

        if query('USE_CYCLOTOMIC_FACTOR'):
            factors = self._zz_cyclotomic_factor(g)

        if factors is None:
            factors = self._zz_zassenhaus(g)

        return cont, _sort_factors(factors, multiple=False)

    def _zz_hensel_lift(self, p, f, f_list, l):
        """
        Multifactor Hensel lifting in `Z[x]`.

        Given a prime `p`, polynomial `f` over `Z[x]` such that `lc(f)`
        is a unit modulo `p`, monic pair-wise coprime polynomials `f_i`
        over `Z[x]` satisfying::

            f = lc(f) f_1 ... f_r (mod p)

        and a positive integer `l`, returns a list of monic polynomials
        `F_1`, `F_2`, ..., `F_r` satisfying::

           f = lc(f) F_1 ... F_r (mod p**l)

           F_i = f_i (mod p), i = 1..r

        References
        ==========

        * :cite:`Gathen1999modern`

        """
        domain = self.domain

        r = len(f_list)
        lc = f.LC

        if r == 1:
            F = f * domain.gcdex(lc, p**l)[0]
            return [F.trunc_ground(p**l)]

        m = p
        k = r // 2
        d = math.ceil(math.log(l, 2))
        p_domain = domain.finite_field(p)

        g = self.ground_new(lc)
        g = g.set_domain(p_domain)

        for f_i in f_list[:k]:
            g *= f_i

        h = f_list[k].set_domain(p_domain)

        for f_i in f_list[k + 1:]:
            h *= f_i

        s, t, _ = self.clone(domain=p_domain).gcdex(g, h)

        g, h, s, t = map(operator.methodcaller('set_domain', domain),
                         (g, h, s, t))

        for _ in range(1, d + 1):
            (g, h, s, t), m = self._zz_hensel_step(m, f, g, h, s, t), m**2

        return (self._zz_hensel_lift(p, g, f_list[:k], l) +
                self._zz_hensel_lift(p, h, f_list[k:], l))

    def _zz_hensel_step(self, m, f, g, h, s, t):
        """
        One step in Hensel lifting in `Z[x]`.

        Given positive integer `m` and `Z[x]` polynomials `f`, `g`, `h`, `s`
        and `t` such that::

            f == g*h (mod m)
            s*g + t*h == 1 (mod m)

            lc(f) is not a zero divisor (mod m)
            lc(h) == 1

            deg(f) == deg(g) + deg(h)
            deg(s) < deg(h)
            deg(t) < deg(g)

        returns polynomials `G`, `H`, `S` and `T`, such that::

            f == G*H (mod m**2)
            S*G + T*H == 1 (mod m**2)

        References
        ==========

        * :cite:`Gathen1999modern`

        """
        M = m**2

        e = f - g*h
        e = e.trunc_ground(M)

        q, r = divmod(s*e, h)

        q = q.trunc_ground(M)
        r = r.trunc_ground(M)

        u = t*e + q*g
        G = (g + u).trunc_ground(M)
        H = (h + r).trunc_ground(M)

        u = s*G + t*H
        b = (u - 1).trunc_ground(M)

        c, d = divmod(s*b, H)

        c = c.trunc_ground(M)
        d = d.trunc_ground(M)

        u = t*b + c*G
        S = (s - d).trunc_ground(M)
        T = (t - u).trunc_ground(M)

        return G, H, S, T

    def factor_list(self, f):
        """Factor polynomials into irreducibles in `K[X]`."""
        domain = self.domain

        cont, f = f.primitive()

        if domain.is_FiniteField:
            if self.is_multivariate:
                raise NotImplementedError('multivariate polynomials over finite fields')

            coeff = f.LC
            f = f.monic()

            factors = []

            for g, n in f.sqf_list()[1]:
                for h in self._gf_factor_sqf(g):
                    factors.append((h, n))
        elif domain.is_AlgebraicField:
            from .factorization_alg_field import efactor

            _factor_aa_methods = {'trager': self._aa_factor_trager,
                                  'modular': efactor}
            method = _factor_aa_methods[query('AA_FACTOR_METHOD')]
            coeff, factors = method(f)
        else:
            if not domain.is_Exact:
                domain_inexact, domain = domain, domain.get_exact()
                f = f.set_domain(domain)
            else:
                domain_inexact = None

            if domain.is_Field:
                domain1 = domain.ring

                denom, f = f.clear_denoms(convert=True)
            else:
                domain1 = domain

            if domain1.is_IntegerRing:
                ring = self.clone(domain=domain1)
                coeff, factors = ring._zz_factor(f)
            elif domain1.is_PolynomialRing:
                f = f.inject()
                ring = f.ring

                coeff, factors = ring.factor_list(f)

                for i, (f, k) in enumerate(factors):
                    factors[i] = (f.eject(*ring.gens[-domain1.ngens:]), k)

                coeff = domain1.convert(coeff)
            else:
                raise DomainError(f'factorization not supported over {domain}')

            if domain.is_Field:
                for i, (f, k) in enumerate(factors):
                    factors[i] = (f.set_domain(domain), k)

                coeff = domain.convert(coeff)
                coeff = domain.quo(coeff, denom)

                if domain_inexact:
                    for i, (f, k) in enumerate(factors):
                        f = f.quo_ground(denom)
                        f = f.set_domain(domain_inexact)
                        factors[i] = (f, k)
                        coeff *= denom**k

                    coeff = domain_inexact.convert(coeff)

        return coeff*cont, _sort_factors(factors)

    def _zz_irreducible_p(self, f):
        """Test irreducibility using Eisenstein's criterion."""
        assert self.is_univariate

        lc = f.LC
        tc = f[1]

        f -= f.leading_term()
        e_fc = f.content()

        if e_fc:
            e_ff = factorint(int(e_fc))

            for p in e_ff:
                if lc % p and tc % p**2:
                    return True

    def _gf_irreducible_p_ben_or(self, f):
        """
        Ben-Or's polynomial irreducibility test over finite fields.

        References
        ==========

        * :cite:`Ben-Or1981ff`

        """
        assert self.is_univariate

        domain = self.domain
        n, q = f.degree(), domain.order

        if n <= 1:
            return True

        x = self.gens[0]
        f = f.monic()

        H = h = pow(x, q, f)

        for _ in range(n//2):
            g = h - x

            if self.gcd(f, g) == 1:
                h = h.compose(x, H) % f
            else:
                return False

        return True

    def _gf_irreducible_p_rabin(self, f):
        """
        Rabin's polynomial irreducibility test over finite fields.

        References
        ==========

        * :cite:`Gathen1999modern`, algorithm 14.36

        """
        assert self.is_univariate

        domain = self.domain
        n, q = f.degree(), domain.order

        if n <= 1:
            return True

        x = self.gens[0]
        f = f.monic()

        indices = {n//d for d in factorint(n)}

        H = h = pow(x, q, f)

        for i in range(1, n):
            if i in indices:
                g = h - x

                if self.gcd(f, g) != 1:
                    return False

            h = h.compose(x, H) % f

        return h == x

    def _cyclotomic_decompose(self, n):
        x = self.gens[0]
        H = [x - 1]

        for p, k in factorint(n).items():
            Q = [h.compose(x, x**p) // h for h in H]
            H.extend(Q)

            for _ in range(1, k):
                Q = [q.compose(x, x**p) for q in Q]
                H.extend(Q)

        return H

    def _zz_cyclotomic_factor(self, f):
        """
        Efficiently factor polynomials `x**n - 1` and `x**n + 1` in `Z[x]`.

        Given a univariate polynomial `f` in `Z[x]` returns a list of factors
        of `f`, provided that `f` is in the form `x**n - 1` or `x**n + 1` for
        `n >= 1`. Otherwise returns None.

        Factorization is performed using cyclotomic decomposition of `f`,
        which makes this method much faster that any other direct factorization
        approach (e.g. Zassenhaus's).

        References
        ==========

        * :cite:`MathWorld-Cyclotomic-Poly`

        """
        lc_f, tc_f = f.LC, f[1]

        if f.is_ground:
            return

        if lc_f != 1 or tc_f not in [-1, 1]:
            return

        if any(bool(cf) for cf in f.all_coeffs()[1:-1]):
            return

        n = f.degree()
        F = self._cyclotomic_decompose(n)

        if tc_f != 1:
            return F
        else:
            H = []

            for h in self._cyclotomic_decompose(2*n):
                if h not in F:
                    H.append(h)

            return H

    def _cyclotomic_p(self, f, irreducible=False):
        """
        Efficiently test if ``f`` is a cyclotomic polynomial.

        Examples
        ========

        >>> _, x = ring('x', ZZ)

        >>> (x**16 + x**14 - x**10 + x**8 - x**6 + x**2 + 1).is_cyclotomic
        False

        >>> (x**16 + x**14 - x**10 - x**8 - x**6 + x**2 + 1).is_cyclotomic
        True

        """
        domain = self.domain

        if domain.is_RationalField:
            try:
                f = f.set_domain(domain.ring)
                return f.is_cyclotomic
            except CoercionFailed:
                return False
        elif not domain.is_IntegerRing:
            return False

        x = self.gens[0]

        lc = f.LC
        tc = f[1]

        if lc != 1 or tc not in (1, -1):
            return False

        if not irreducible:
            coeff, factors = f.factor_list()

            if coeff != 1 or factors != [(f, 1)]:
                return False

        n = f.degree()
        g, h = self.zero, self.zero

        for j, i in enumerate(range(n, -1, -2)):
            g += f[(i,)]*x**j

        for j, i in enumerate(range(n - 1, -1, -2)):
            h += f[(i,)]*x**j

        g = g**2
        h = h**2

        F = g - h*self.from_terms([((1,), domain.one)])

        if F.LC < 0:
            F = -F

        if F == f:
            return True

        g = f.compose(x, -x)

        if g.LC < 0:
            g = -g

        if F == g and g.is_cyclotomic:
            return True

        G = F.sqf_part()

        if G**2 == F and G.is_cyclotomic:
            return True

        return False

    def _zz_cyclotomic_poly(self, n):
        """Efficiently generate n-th cyclotomic polynomial."""
        x = self.gens[0]

        h = x - 1

        for p, k in factorint(n).items():
            h = h.compose(x, x**p) // h
            h = h.compose(x, x**(p**(k - 1)))

        return h

    def _univar_zz_diophantine(self, F, m, p):
        """Wang/EEZ: Solve univariate Diophantine equations."""
        domain = self.domain
        m = self.from_terms([((m,), domain.one)])

        if len(F) == 2:
            p_domain = domain.finite_field(p)
            p_ring = self.clone(domain=p_domain)
            f, g = map(operator.methodcaller('set_domain', p_domain), F)

            s, t, _ = p_ring.gcdex(g, f)

            s *= m
            t *= m

            q, s = divmod(s, f)
            s = s.set_domain(domain)

            t += q*g
            t = t.set_domain(domain)

            result = [s, t]
        else:
            G = [F[-1]]

            for f in reversed(F[1:-1]):
                G.insert(0, f*G[0])

            S, T = [], [self.one]

            for f, g in zip(F, G):
                t, s = self._zz_diophantine([g, f], T[-1], [], 0, p)
                T.append(t)
                S.append(s)

            result, S = [], S + [T[-1]]
            p_domain = domain.finite_field(p)

            for s, f in zip(S, F):
                s *= m
                s, f = map(operator.methodcaller('set_domain', p_domain),
                           (s, f))
                s = (s % f).set_domain(domain)

                result.append(s)

        return result

    def _zz_diophantine(self, F, c, A, d, p):
        """Wang/EEZ: Solve multivariate Diophantine equations."""
        domain = self.domain

        if not A:
            S = [self.zero for _ in F]
            n = c.degree()

            for i, coeff in enumerate(c.all_coeffs()):
                if not coeff:
                    continue

                T = self._univar_zz_diophantine(F, i, p)

                for j, (s, t) in enumerate(zip(S, T)):
                    t *= coeff
                    S[j] = (s + t).trunc_ground(p)
        else:
            n = len(A)
            e = functools.reduce(operator.mul, F)

            a, A = A[-1], A[:-1]
            B, G = [], []

            for f in F:
                B.append(e // f)
                G.append(f.eval(x=n, a=a))

            C = c.eval(x=n, a=a)

            S = self.drop(n)._zz_diophantine(G, C, A, d, p)
            S = [s.set_ring(self) for s in S]

            for s, b in zip(S, B):
                c -= s*b

            c = c.trunc_ground(p)

            m = self.gens[n] - a
            M = self.one

            for k in range(d):
                k = domain(k)
                if not c:
                    break

                M *= m
                C = c.diff(x=n, m=int(k + 1)).eval(x=n, a=a)

                if C:
                    C = C.quo_ground(domain.factorial(k + 1))
                    T = C.ring._zz_diophantine(G, C, A, d, p)

                    for i, t in enumerate(T):
                        T[i] = t.set_ring(self)*M

                    for i, (s, t) in enumerate(zip(S, T)):
                        S[i] = s + t

                    for t, b in zip(T, B):
                        c -= t*b

                    c = c.trunc_ground(p)

            S = [s.trunc_ground(p) for s in S]

        return S

    def _zz_wang(self, f, mod=None, seed=None):
        """
        Factor primitive square-free polynomials in `Z[X]`.

        Given a multivariate polynomial `f` in `Z[x_1,...,x_n]`, which is
        primitive and square-free in `x_1`, computes factorization of `f` into
        irreducibles over integers.

        The procedure is based on Wang's Enhanced Extended Zassenhaus
        algorithm. The algorithm works by viewing `f` as a univariate polynomial
        in `Z[x_2,...,x_n][x_1]`, for which an evaluation mapping is computed::

                          x_2 -> a_2, ..., x_n -> a_n

        where `a_i`, for `i = 2, ..., n`, are carefully chosen integers.  The
        mapping is used to transform `f` into a univariate polynomial in `Z[x_1]`,
        which can be factored efficiently using Zassenhaus algorithm. The last
        step is to lift univariate factors to obtain true multivariate
        factors. For this purpose a parallel Hensel lifting procedure is used.

        The parameter ``seed`` is passed to _randint and can be used to seed randint
        (when an integer) or (for testing purposes) can be a sequence of numbers.

        References
        ==========

        * :cite:`Wang1978improved`
        * :cite:`Geddes1992algorithms`

        """
        from ..utilities.randtest import _randint

        randint = _randint(seed)

        ct, T = f.eject(*self.gens[1:]).LC.factor_list()

        domain = self.domain
        uring = self.drop(*self.gens[1:])
        b = self._zz_mignotte_bound(f)
        p = domain(nextprime(b))

        if mod is None:
            if self.ngens == 2:
                mod = 2
            else:
                mod = 1

        history, configs, A, r = set(), [], [domain.zero]*(self.ngens - 1), None

        try:
            cs, s, E = self._zz_wang_test_points(f, T, ct, A)

            _, H = uring._zz_factor_sqf(s)

            r = len(H)

            if r == 1:
                return [f]

            configs = [(s, cs, E, H, A)]
        except EvaluationFailed:
            pass

        eez_num_configs = query('EEZ_NUMBER_OF_CONFIGS')
        eez_num_tries = query('EEZ_NUMBER_OF_TRIES')
        eez_mod_step = query('EEZ_MODULUS_STEP')

        while len(configs) < eez_num_configs:
            for _ in range(eez_num_tries):
                A = [domain(randint(-mod, mod)) for _ in range(self.ngens - 1)]

                if tuple(A) in history:
                    continue
                history.add(tuple(A))

                try:
                    cs, s, E = self._zz_wang_test_points(f, T, ct, A)
                except EvaluationFailed:
                    continue

                _, H = uring._zz_factor_sqf(s)

                rr = len(H)

                if r is not None:
                    if rr != r:
                        if rr >= r:
                            continue
                        configs, r = [], rr
                else:
                    r = rr

                if r == 1:
                    return [f]

                configs.append((s, cs, E, H, A))

                if len(configs) == eez_num_configs:
                    break
            else:
                mod += eez_mod_step

        s_norm, s_arg, i = None, 0, 0

        for s, _, _, _, _ in configs:
            _s_norm = s.max_norm()

            if s_norm is not None:
                if _s_norm < s_norm:
                    s_norm = _s_norm
                    s_arg = i
            else:
                s_norm = _s_norm

            i += 1

        _, cs, E, H, A = configs[s_arg]
        orig_f = f

        try:
            f, H, LC = self._zz_wang_lead_coeffs(f, T, cs, E, H, A)
            factors = self._zz_wang_hensel_lifting(f, H, LC, A, p)
        except ExtraneousFactors as exc:
            if query('EEZ_RESTART_IF_NEEDED'):
                return self._zz_wang(orig_f, mod + 1)
            else:
                raise ExtraneousFactors('we need to restart algorithm '
                                        'with better parameters') from exc

        result = []

        for f in factors:
            _, f = f.primitive()

            result.append(f)

        return result

    def _zz_wang_test_points(self, f, T, ct, A):
        """Wang/EEZ: Test evaluation points for suitability."""
        if not f.eject(*self.gens[1:]).LC(*A):
            raise EvaluationFailed('no luck')

        g = f.eject(0)(*A)

        if not g.is_squarefree:
            raise EvaluationFailed('no luck')

        c, h = g.primitive()

        E = [t(*A) for t, _ in T]
        D = self._zz_wang_non_divisors(E, c, ct)

        if D is not None:
            return c, h, E
        else:
            raise EvaluationFailed('no luck')

    def _zz_wang_non_divisors(self, E, cs, ct):
        """Wang/EEZ: Compute a set of valid divisors."""
        domain = self.domain
        result = [cs*ct]

        for q in E:
            q = abs(q)

            for r in reversed(result):
                while r != 1:
                    r = domain.gcd(r, q)
                    q = q // r

                if q == 1:
                    return

            result.append(q)

        return result[1:]

    def _zz_wang_lead_coeffs(self, f, T, cs, E, H, A):
        """Wang/EEZ: Compute correct leading coefficients."""
        domain = self.domain
        c_ring = self.drop(0)
        C, J = [], [0]*len(E)

        for h in H:
            c = c_ring.one
            d = h.LC*cs

            for i in reversed(range(len(E))):
                k, e, t = 0, E[i], T[i][0]

                while not d % e:
                    d, k = d//e, k + 1

                if k != 0:
                    c *= t**k
                    J[i] = 1

            C.append(c)

        if any(not j for j in J):  # pragma: no cover
            raise ExtraneousFactors

        CC, HH = [], []

        for c, h in zip(C, H):
            d = c(*A)
            lc = h.LC

            if cs == 1:
                cc = lc//d
            else:
                g = domain.gcd(lc, d)
                d //= g
                cc = lc//g
                h *= d
                cs //= d

            c *= cc

            CC.append(c)
            HH.append(h)

        if cs == 1:
            return f, HH, CC

        CCC, HHH = [], []

        for c, h in zip(CC, HH):
            CCC.append(c*cs)
            HHH.append(h*cs)

        f *= cs**(len(H) - 1)

        return f, HHH, CCC

    def _zz_wang_hensel_lifting(self, f, H, LC, A, p):
        """Wang/EEZ: Parallel Hensel lifting algorithm."""
        domain = self.domain
        S, n = [f], len(A)

        H = list(H)

        for i, a in enumerate(reversed(A[1:])):
            s = S[0].eval(x=n - i, a=a)
            S.insert(0, s.trunc_ground(p))

        d = max(f.degree(x) for x in self.gens[1:])

        for j, s, a in zip(range(2, n + 2), S, A):
            G, w = list(H), j - 1
            s_ring = s.ring

            I, J = A[:w - 1], A[w:]

            for i, (h, lc) in enumerate(zip(H, LC)):
                if J:
                    lc = lc.eject(*lc.ring.gens[:-len(J)])(*J)
                lc = lc.trunc_ground(p)
                h, lc = map(operator.methodcaller('set_ring', s_ring), (h, lc))
                lt = h.eject(*s_ring.gens[1:]).leading_term().inject()
                H[i] = lc*s_ring.gens[0]**lt.degree() + h - lt

            m = s_ring.gens[-1] - a
            M = s_ring.one

            c = functools.reduce(operator.mul, H)
            c = s - c

            dj = s.degree(x=w)

            for k in range(dj):
                k = domain(k)
                if not c:
                    break

                M *= m
                C = c.diff(x=w, m=int(k + 1)).eval(x=w, a=a)

                if C:
                    C = C.quo_ground(domain.factorial(k + 1))
                    T = C.ring._zz_diophantine(G, C, I, d, p)

                    for i, (h, t) in enumerate(zip(H, T)):
                        t = t.set_ring(s_ring)
                        h += t*M
                        H[i] = h.trunc_ground(p)

                    h = functools.reduce(operator.mul, H)
                    h = s - h
                    c = h.trunc_ground(p)

        if functools.reduce(operator.mul, H) != f:
            raise ExtraneousFactors
        return H

    def _gf_Qmatrix(self, f):
        """
        Calculate Berlekamp's ``Q`` matrix.

        Examples
        ========

        >>> R, x = ring('x', FF(5))

        >>> f = 3*x**2 + 2*x + 4
        >>> R._gf_Qmatrix(f)
        [[1, 0], [3, 4]]

        >>> f = x**4 + 1
        >>> R._gf_Qmatrix(f)
        [[1, 0, 0, 0], [0, 4, 0, 0], [0, 0, 1, 0], [0, 0, 0, 4]]

        References
        ==========

        * :cite:`Geddes1992algorithms`, algorithm 8.5

        """
        domain = self.domain
        n, q = f.degree(), domain.order

        r = [domain.one] + [domain.zero]*(n - 1)
        Q = [r.copy()] + [[]]*(n - 1)
        f = f.all_coeffs()

        for i in range(1, (n - 1)*q + 1):
            c, r[1:], r[0] = r[-1], r[:-1], domain.zero
            for j in range(n):
                r[j] -= c*f[j]

            if not i % q:
                Q[i//q] = r.copy()

        return Q

    def _gf_berlekamp(self, f):
        """
        Factor a square-free polynomial over finite fields of small order.

        Examples
        ========

        >>> R, x = ring('x', FF(5))

        >>> R._gf_berlekamp(x**4 + 1)
        [x**2 + 2, x**2 + 3]

        References
        ==========

        * :cite:`Geddes1992algorithms`, algorithm 8.4
        * :cite:`Knuth1985seminumerical`, section 4.6.2

        """
        from .solvers import RawMatrix

        assert self.is_univariate

        domain = self.domain

        Q = self._gf_Qmatrix(f)
        Q = RawMatrix(Q) - RawMatrix.eye(len(Q))
        V = Q.T.nullspace()

        for i, v in enumerate(V):
            V[i] = self.from_list(v)

        factors = [f]

        for v in V[1:]:
            for f in list(factors):
                for s in range(domain.order):
                    h = v - domain(s)
                    g = self.gcd(f, h)

                    if g not in (1, f):
                        factors.remove(f)

                        f //= g
                        factors.extend([f, g])

                    if len(factors) == len(V):
                        return _sort_factors(factors, multiple=False)

        return _sort_factors(factors, multiple=False)

    def _gf_ddf_zassenhaus(self, f):
        """
        Cantor-Zassenhaus: Deterministic Distinct Degree Factorization.

        Given a monic square-free polynomial ``f`` in ``GF(q)[x]``, computes
        partial distinct degree factorization ``f_1 ... f_d`` of ``f`` where
        ``deg(f_i) != deg(f_j)`` for ``i != j``. The result is returned as a
        list of pairs ``(f_i, e_i)`` where ``deg(f_i) > 0`` and ``e_i > 0``
        is an argument to the equal degree factorization routine.

        Examples
        ========

        >>> R, x = ring('x', FF(11))
        >>> R._gf_ddf_zassenhaus(x**15 - 1)
        [(x**5 + 10, 1), (x**10 + x**5 + 1, 2)]

        To obtain factorization into irreducibles, use equal degree factorization
        procedure (EDF) with each of the factors.

        References
        ==========

        * :cite:`Gathen1999modern`, algorithm 14.3
        * :cite:`Geddes1992algorithms`, algorithm 8.8

        See Also
        ========

        _gf_edf_zassenhaus

        """
        domain = self.domain

        factors, q = [], domain.order
        g, x = [self.gens[0]]*2

        for i in range(1, f.degree()//2 + 1):
            g = pow(g, q, f)
            h = self.gcd(f, g - x)

            if h != 1:
                factors.append((h, i))

                f //= h
                g %= f

        if f != 1:
            factors += [(f, f.degree())]

        return factors

    def _gf_edf_zassenhaus(self, f, n):
        """
        Cantor-Zassenhaus: Probabilistic Equal Degree Factorization.

        Given a monic square-free polynomial ``f`` in ``GF(q)[x]`` and
        an integer ``n``, such that ``n`` divides ``deg(f)``, returns all
        irreducible factors ``f_1,...,f_d`` of ``f``, each of degree ``n``.
        EDF procedure gives complete factorization over Galois fields.

        Examples
        ========

        >>> R, x = ring('x', FF(5))
        >>> R._gf_edf_zassenhaus(x**3 + x**2 + x + 1, 1)
        [x + 1, x + 2, x + 3]

        References
        ==========

        * :cite:`Geddes1992algorithms`, algorithm 8.9

        See Also
        ========

        _gf_ddf_zassenhaus

        """
        factors = [f]
        d = f.degree()

        if d <= n:
            return factors

        domain = self.domain
        p, q = domain.characteristic, domain.order
        N = d // n

        while len(factors) < N:
            r = self._gf_random(2*n - 1)

            if p == 2:
                h = r

                for _ in range(1, n):
                    h += pow(r, q, f)
            else:
                h = pow(r, (q**n - 1)//2, f)
                h -= 1

            g = self.gcd(f, h)

            if g not in (1, f):
                factors = (self._gf_edf_zassenhaus(g, n) +
                           self._gf_edf_zassenhaus(f // g, n))

        return _sort_factors(factors, multiple=False)

    def _gf_zassenhaus(self, f):
        """
        Factor a square-free polynomial over finite fields of medium order.

        Examples
        ========

        >>> R, x = ring('x', FF(5))
        >>> R._gf_zassenhaus(x**2 + 4*x + 3)
        [x + 1, x + 3]

        """
        assert self.is_univariate

        factors = []

        for factor, n in self._gf_ddf_zassenhaus(f):
            factors += self._gf_edf_zassenhaus(factor, n)

        return _sort_factors(factors, multiple=False)

    def _gf_ddf_shoup(self, f):
        """
        Kaltofen-Shoup: Deterministic Distinct Degree Factorization.

        Given a monic square-free polynomial ``f`` in ``GF(q)[x]``, computes
        partial distinct degree factorization ``f_1 ... f_d`` of ``f`` where
        ``deg(f_i) != deg(f_j)`` for ``i != j``. The result is returned as a
        list of pairs ``(f_i, e_i)`` where ``deg(f_i) > 0`` and ``e_i > 0``
        is an argument to the equal degree factorization routine.

        Notes
        =====

        This algorithm is an improved version of Zassenhaus algorithm for
        large ``deg(f)`` and order ``q`` (especially for ``deg(f) ~ lg(q)``).

        Examples
        ========

        >>> R, x = ring('x', FF(3))
        >>> R._gf_ddf_shoup(x**6 - x**5 + x**4 + x**3 - x)
        [(x**2 + x, 1), (x**4 + x**3 + x + 2, 2)]

        References
        ==========

        * :cite:`Kaltofen1998subquadratic`, algorithm D
        * :cite:`Shoup1995factor`
        * :cite:`Gathen1992frobenious`

        See Also
        ========

        _gf_edf_shoup

        """
        domain = self.domain

        n, q = f.degree(), domain.order
        k = math.isqrt(n//2 - 1) + 1 if n > 1 else 0
        x = self.gens[0]

        h = pow(x, q, f)

        # U[i] = x**(q**i)
        U = [x, h] + [self.zero]*(k - 1)

        for i in range(2, k + 1):
            U[i] = U[i - 1].compose(x, h) % f

        h, U = U[k], U[:k]
        # V[i] = x**(q**(k*(i+1)))
        V = [h] + [self.zero]*(k - 1)

        for i in range(1, k):
            V[i] = V[i - 1].compose(x, h) % f

        factors = []

        for i, v in enumerate(V):
            h, j = self.one, k - 1

            for u in U:
                g = v - u
                h *= g
                h %= f

            g = self.gcd(f, h)
            f //= g

            for u in reversed(U):
                h = v - u
                F = self.gcd(g, h)

                if F != 1:
                    factors.append((F, k*(i + 1) - j))

                g //= F
                j -= 1

        if f != 1:
            factors.append((f, f.degree()))

        return factors

    def _gf_trace_map(self, a, b, c, n, f):
        """
        Compute polynomial trace map in ``GF(q)[x]/(f)``.

        Given a polynomial ``f`` in ``GF(q)[x]``, polynomials ``a``, ``b``,
        ``c`` in the quotient ring ``GF(q)[x]/(f)`` such that ``b = c**t
        (mod f)`` for some positive power ``t`` of ``q``, and a positive
        integer ``n``, returns a mapping::

           a -> a**t**n, a + a**t + a**t**2 + ... + a**t**n (mod f)

        In factorization context, ``b = x**q mod f`` and ``c = x mod f``.
        This way we can efficiently compute trace polynomials in equal
        degree factorization routine, much faster than with other methods,
        like iterated Frobenius algorithm, for large degrees.

        Examples
        ========

        >>> R, x = ring('x', FF(5))
        >>> a = x + 2
        >>> b = 4*x + 4
        >>> c = x + 1
        >>> f = 3*x**2 + 2*x + 4
        >>> R._gf_trace_map(a, b, c, 4, f)
        (x + 3, x + 3)

        References
        ==========

        * :cite:`Gathen1992ComputingFM`, algorithm 5.2

        """
        u = a.compose(0, b) % f
        v = b

        if n & 1:
            U = a + u
            V = b
        else:
            U = a
            V = c

        n >>= 1

        while n:
            u += u.compose(0, v) % f
            v = v.compose(0, v) % f

            if n & 1:
                U += u.compose(0, V) % f
                V = v.compose(0, V) % f

            n >>= 1

        return a.compose(0, V) % f, U

    def _gf_edf_shoup(self, f, n):
        """
        Gathen-Shoup: Probabilistic Equal Degree Factorization.

        Given a monic square-free polynomial ``f`` in ``GF(q)[x]`` and
        an integer ``n``, such that ``n`` divides ``deg(f)``, returns all
        irreducible factors ``f_1,...,f_d`` of ``f``, each of degree ``n``.
        EDF procedure gives complete factorization over Galois fields.

        Notes
        =====

        This algorithm is an improved version of Zassenhaus algorithm for
        large ``deg(f)`` and order ``q`` (especially for ``deg(f) ~ lg(q)``).

        Examples
        ========

        >>> R, x = ring('x', FF(2917))
        >>> R._gf_edf_shoup(x**2 + 2837*x + 2277, 1)
        [x + 852, x + 1985]

        References
        ==========

        * :cite:`Shoup1991ffactor`
        * :cite:`Gathen1992ComputingFM`, algorithm 3.6

        See Also
        ========

        _gf_ddf_shoup

        """
        domain = self.domain
        q, p = domain.order, domain.characteristic
        N = f.degree()

        if not N:
            return []
        if N <= n:
            return [f]

        factors, x = [f], self.gens[0]

        r = self._gf_random(N - 1)

        h = pow(x, q, f)
        H = self._gf_trace_map(r, h, x, n - 1, f)[1]

        if p == 2:
            h1 = self.gcd(f, H)
            h2 = f // h1

            factors = self._gf_edf_shoup(h1, n) + self._gf_edf_shoup(h2, n)
        else:
            h = pow(H, (q - 1)//2, f)

            h1 = self.gcd(f, h)
            h2 = self.gcd(f, h - 1)
            h3 = f // (h1 * h2)

            factors = (self._gf_edf_shoup(h1, n) + self._gf_edf_shoup(h2, n) +
                       self._gf_edf_shoup(h3, n))

        return _sort_factors(factors, multiple=False)

    def _gf_shoup(self, f):
        """
        Factor a square-free polynomial over finite fields of large order.

        Examples
        ========

        >>> R, x = ring('x', FF(5))
        >>> R._gf_shoup(x**2 + 4*x + 3)
        [x + 1, x + 3]

        """
        assert self.is_univariate

        factors = []

        for factor, n in self._gf_ddf_shoup(f):
            factors += self._gf_edf_shoup(factor, n)

        return _sort_factors(factors, multiple=False)

    def _gf_factor_sqf(self, f):
        """
        Factor a square-free polynomial ``f`` in ``GF(q)[x]``.

        Returns its complete factorization into irreducibles::

                     f_1(x) f_2(x) ... f_d(x)

        where each ``f_i`` is a monic polynomial and ``gcd(f_i, f_j) == 1``,
        for ``i != j``.  The result is given as a list of factors of ``f``.

        Square-free factors of ``f`` can be factored into irreducibles over
        finite fields using three very different methods:

        Berlekamp
            efficient for very small values of order ``q`` (usually ``q < 25``)
        Cantor-Zassenhaus
            efficient on average input and with "typical" ``q``
        Shoup-Kaltofen-Gathen
            efficient with very large inputs and order

        If you want to use a specific factorization method - set
        ``GF_FACTOR_METHOD`` configuration option with one of ``"berlekamp"``,
        ``"zassenhaus"`` or ``"shoup"`` values.

        Examples
        ========

        >>> R, x = ring('x', FF(5))
        >>> f = x**2 + 4*x + 3
        >>> R._gf_factor_sqf(f)
        [x + 1, x + 3]

        References
        ==========

        * :cite:`Gathen1999modern`, chapter 14

        """
        _factor_methods = {
            'berlekamp': self._gf_berlekamp,  # ``p`` : small
            'zassenhaus': self._gf_zassenhaus,  # ``p`` : medium
            'shoup': self._gf_shoup,      # ``p`` : large
        }
        method = query('GF_FACTOR_METHOD')

        return _factor_methods[method](f)



from ..core import Dummy, Expr, Float, PoleError, Rational, nan, oo, sympify
from ..core.function import UndefinedFunction
from ..functions import cos, sign, sin
from ..sets import Reals
from .gruntz import limitinf
from .order import Order


def limit(expr, z, z0, dir=None):
    """
    Compute the directional limit of ``expr`` at the point ``z0``.

    Examples
    ========

    >>> limit(1/x, x, 0)
    oo
    >>> limit(1/x, x, 0, dir=1)
    -oo
    >>> limit(1/x, x, oo)
    0

    See Also
    ========

    Limit

    """
    return Limit(expr, z, z0, dir).doit(deep=False)


def heuristics(e, z, z0, dir):
    e = e.expand()
    if (e.is_Mul or e.is_Add or e.is_Pow or
            (e.is_Function and not e.is_Piecewise and
             not isinstance(e.func, UndefinedFunction))):
        r = []
        for a in e.args:
            l = limit(a, z, z0, dir)
            if (l.has(oo) and (l.func not in (sin, cos) and
                               l.is_finite is None)) or isinstance(l, Limit):
                return
            r.append(l)

        if (rv := e.func(*r)) is not nan:
            return rv


class Limit(Expr):
    r"""
    Represents an unevaluated directional limit of ``expr`` at the point ``z0``.

    Parameters
    ==========

    expr : Expr
        algebraic expression
    z    : Symbol
        variable of the ``expr``
    z0   : Expr
        limit point, `z_0`
    dir  : Expr or Reals, optional
        selects the direction (as ``sign(dir)``) to approach the limit point
        if the ``dir`` is an Expr.  For infinite ``z0``, the default value
        is determined from the direction of the infinity (e.g., the limit
        from the left, ``dir=1``, for ``oo``).  Otherwise, the default is
        the limit from the right, ``dir=-1``.   If ``dir=Reals``, the limit
        is the bidirectional real limit.

    Examples
    ========

    >>> Limit(1/x, x, 0, dir=1)
    Limit(1/x, x, 0, dir=1)
    >>> _.doit()
    -oo

    See Also
    ========

    limit

    """

    def __new__(cls, e, z, z0, dir=None):
        e, z, z0, dir = map(sympify, [e, z, z0, dir])

        if z0.is_infinite:
            dir = sign(z0).simplify()
        elif dir is None:
            dir = Rational(-1)

        if dir == Reals:
            pass
        elif isinstance(dir, Expr) and dir.is_nonzero:
            dir = dir/abs(dir)
        else:
            raise ValueError('direction must be either a nonzero expression '
                             f'or Reals, not {dir}')

        obj = super().__new__(cls)
        obj._args = (e, z, z0, dir)
        return obj

    @property
    def free_symbols(self):
        e, z, z0 = self.args[:3]
        return (e.free_symbols - z.free_symbols) | z0.free_symbols

    def doit(self, **hints):
        """
        Evaluates limit.

        Notes
        =====

        First we handle some trivial cases (i.e. constant), then try
        Gruntz algorithm (see the :py:mod:`~diofant.series.gruntz` module).

        """
        e, z, z0, dir = self.args

        if hints.get('deep', True):
            e = e.doit(**hints)
            z = z.doit(**hints)
            z0 = z0.doit(**hints)

        if dir == Reals:
            right = limit(e, z, z0)
            left = limit(e, z, z0, 1)
            if not left.equals(right):
                raise PoleError(f'left and right limits for the expression {e} '
                                f'at point {z}={z0} seems to be not equal')
            return right

        if (has_Floats := e.has(Float) or z0.has(Float)):
            e = e.subs({k: Rational(k) for k in e.atoms(Float)})
            z0 = z0.subs({k: Rational(k) for k in z0.atoms(Float)})

        if z0.has(z):
            newz = z.as_dummy()
            r = limit(e.subs({z: newz}), newz, z0, dir)
            if isinstance(r, Limit):
                r = r.subs({newz: z})
            return r

        # Use a fresh variable to remove assumptions on the dummy variable.
        newz = Dummy('z')
        e, z = e.subs({z: newz}), newz

        if not e.has(z):
            return e

        if e.has(Order) and (order := e.getO()) and (z, z0) in order.args[1:]:
            order = limit(order.expr, z, z0, dir)
            e = e.removeO() + order

        # Convert to the limit z->oo and use Gruntz algorithm.
        e = e.subs({z: dir*z})
        z0 = z0/dir
        if z0 != oo:
            e = e.subs({z: z0 - 1/z})

        # We need a fresh variable with correct assumptions.
        newz = Dummy(z.name, positive=True, finite=True)
        e, z = e.subs({z: newz}), newz

        if e.is_Boolean or e.is_Relational:
            try:
                has_oo = e.as_set().closure.contains(oo)
            except NotImplementedError:
                return self
            if has_oo.is_Boolean:
                return has_oo
            raise NotImplementedError

        try:
            r = limitinf(e, z)
        except (PoleError, ValueError, NotImplementedError):
            r = None
            if hints.get('heuristics', True):
                r = heuristics(*self.args)
            if r is None:
                return self

        if has_Floats:
            r = r.evalf()

        return r



"""
This test suite is testing the Gruntz algorithm implementation.

See the documentation in the gruntz module.  The
algorithm itself is highly recursive by nature, so ``compare()`` is
logically the lowest part of the algorithm, yet in some sense it's the most
complex part, because it needs to calculate a limit to return the result.
"""

import pytest

from diofant import (Add, E, Ei, EulerGamma, GoldenRatio, I, Integer, Li,
                     Limit, Max, Min, Mul, Pow, Rational, Symbol, acosh, acot,
                     airyai, airybi, atan, binomial, cbrt, cos, cosh, coth,
                     digamma, erf, exp, factorial, fibonacci, gamma, li, limit,
                     log, loggamma, oo, pi, root, sign, sin, sinh, sqrt, tan,
                     tanh, zeta)
from diofant.abc import a, n, y
from diofant.series.gruntz import compare, leadterm, mrv, rewrite, signinf


__all__ = ()

x = Symbol('x', real=True, positive=True)
m = Symbol('m', real=True, positive=True)


@pytest.mark.slow
def test_gruntz_evaluation():
    # Gruntz' thesis, problems 8.1-20, pp. 122 to 123
    assert limit(exp(x)*(exp(1/x - exp(-x)) - exp(1/x)), x, oo) == -1
    assert limit(exp(x)*(exp(1/x + exp(-x) + exp(-x**2))
                         - exp(1/x - exp(-exp(x)))), x, oo) == 1
    assert limit(exp(exp(x - exp(-x))/(1 - 1/x)) - exp(exp(x)), x, oo) == oo
    assert limit(exp(exp(exp(x)/(1 - 1/x)))
                 - exp(exp(exp(x)/(1 - 1/x - log(x)**(-log(x))))), x, oo) == -oo
    assert limit(exp(exp(exp(x + exp(-x)))) / exp(exp(exp(x))), x, oo) == oo
    assert limit(exp(exp(exp(x))) / exp(exp(exp(x - exp(-exp(x))))),
                 x, oo) == oo
    assert limit(exp(exp(exp(x))) / exp(exp(exp(x - exp(-exp(exp(x)))))),
                 x, oo) == 1
    assert limit(exp(exp(x)) / exp(exp(x - exp(-exp(exp(x))))), x, oo) == 1
    assert limit(log(x)**2*exp(sqrt(log(x))*(log(log(x)))**2 *
                 exp(sqrt(log(log(x)))*(log(log(log(x))))**3))/sqrt(x),
                 x, oo) == 0
    assert limit((x*log(x)*(log(x*exp(x) - x**2))**2) /
                 (log(log(x**2 + 2*exp(exp(3*x**3*log(x)))))),
                 x, oo) == Rational(1, 3)
    assert limit((exp(x*exp(-x)/(exp(-x) + exp(-2*x**2/(x + 1)))) -
                 exp(x))/x, x, oo) == -exp(2)
    assert limit((3**x + 5**x)**(1/x), x, oo) == 5
    assert limit(x/log(x**(log(x**(log(2)/log(x))))), x, oo) == oo
    assert limit(exp(exp(2*log(x**5 + x)*log(log(x)))) /
                 exp(exp(10*log(x)*log(log(x)))), x, oo) == oo
    assert limit(exp(exp(Rational(5, 2)*x**(-Rational(5, 7)) +
                         Rational(21, 8)*x**Rational(6, 11) + 2*x**(-8) +
                         Rational(54, 17)*x**Rational(49, 45)))**8 /
                 log(log(-log(Rational(4, 3) *
                     x**(-Rational(5, 14)))))**Rational(7, 6), x, oo) == oo
    assert limit((exp(4*x*exp(-x)/(1/exp(x) + 1/exp(2*x**2/(x + 1)))) -
                  exp(x))/exp(x)**4, x, oo) == 1
    assert limit(exp(x*exp(-x)/(exp(-x) +
                                exp(-2*x**2/(x + 1))))/exp(x), x, oo) == 1
    assert limit((exp(exp(-x/(1 + exp(-x)))) *
                  exp(-x/(1 + exp(-x/(1 + exp(-x))))) *
                  exp(exp(-x + exp(-x/(1 + exp(-x)))))) /
                 (exp(-x/(1 + exp(-x))))**2 - exp(x) + x, x, oo) == 2
    assert limit(log(x)*(log(log(x) + log(log(x))) - log(log(x))) /
                 (log(log(x) + log(log(log(x))))), x, oo) == 1
    assert limit(exp((log(log(x + exp(log(x)*log(log(x)))))) /
                     log(log(log(exp(x) + x + log(x))))), x, oo) == E
    # Another
    assert limit(exp(exp(exp(x + exp(-x))))/exp(exp(x)), x, oo) == oo
    assert limit(log(log(x*exp(x*exp(x)) + 1)) -
                 exp(exp(log(log(x)) + 1/x)), x, oo) == 0


def test_gruntz_eval_special():
    # Gruntz, p. 126
    assert limit(exp(x)*(sin(1/x + exp(-x)) - sin(1/x + exp(-x**2))),
                 x, oo) == 1
    assert limit((erf(x - exp(-exp(x))) - erf(x))*exp(exp(x))*exp(x**2),
                 x, oo) == -2/sqrt(pi)
    assert limit(exp(exp(x))*(exp(sin(1/x + exp(-exp(x)))) -
                              exp(sin(1/x))), x, oo) == 1
    assert limit(exp(x)*(gamma(x + exp(-x)) - gamma(x)), x, oo) == oo
    assert limit(exp(exp(digamma(digamma(x))))/x, x, oo) == exp(-Rational(1, 2))
    assert limit(exp(exp(digamma(log(x))))/x, x, oo) == exp(-Rational(1, 2))
    assert limit(digamma(digamma(digamma(x))), x, oo) == oo
    assert limit(loggamma(loggamma(x)), x, oo) == oo
    assert limit(((gamma(x + 1/gamma(x)) - gamma(x))/log(x) - cos(1/x)) *
                 x*log(x), x, oo) == -Rational(1, 2)
    assert limit(x*(gamma(x - 1/gamma(x)) - gamma(x) + log(x)),
                 x, oo) == Rational(1, 2)
    assert limit((gamma(x + 1/gamma(x)) - gamma(x))/log(x), x, oo) == 1
    assert limit(erf(log(1/x)), x, oo) == -1


@pytest.mark.slow
def test_gruntz_eval_special_slow():
    assert limit(gamma(x + 1)/sqrt(2*pi) - exp(-x)*(x**(x + Rational(1, 2)) +
                 x**(x - Rational(1, 2))/12), x, oo) == oo
    assert limit(exp(exp(exp(digamma(digamma(digamma(x))))))/x, x, oo) == 0
    assert limit(exp(gamma(x - exp(-x))*exp(1/x)) - exp(gamma(x)), x, oo) == oo
    assert limit((Ei(x - exp(-exp(x))) - Ei(x)) *
                 exp(-x)*exp(exp(x))*x, x, oo) == -1
    assert limit(exp((log(2) + 1)*x)*(zeta(x + exp(-x)) - zeta(x)),
                 x, oo) == -log(2)
    # TODO 8.36 (bessel)
    assert limit(Max(x, exp(x))/log(Min(exp(-x), exp(-exp(x)))), x, oo) == -1


def test_gruntz_other():
    assert limit(sqrt(log(x + 1)) - sqrt(log(x)), x, oo) == 0  # p12, 2.5
    assert limit(((1 + 1/x)**y - 1)*x, x, oo) == y  # p12, 2.6
    assert limit(x**n/exp(x), x, oo) == 0  # p14, 2.9
    assert limit((1 + 1/x)*x - 1/log(1 + 1/x),
                 x, oo) == Rational(1, 2)  # p15, 2.10
    assert limit((root(1 + 1/x, n) - 1)/(root(1 + 1/x, m) - 1),
                 x, oo) == m/n  # p13, 2.7


def test_gruntz_hyperbolic():
    assert limit(cosh(x), x, oo) == oo
    assert limit(cosh(-x), x, oo) == oo
    assert limit(sinh(x), x, oo) == oo
    assert limit(sinh(-x), x, oo) == -oo
    assert limit(2*cosh(x)*exp(x), x, oo) == oo
    assert limit(2*cosh(-x)*exp(-x), x, oo) == 1
    assert limit(2*sinh(x)*exp(x), x, oo) == oo
    assert limit(2*sinh(-x)*exp(-x), x, oo) == -1
    assert limit(tanh(x), x, oo) == 1
    assert limit(tanh(-x), x, oo) == -1
    assert limit(coth(x), x, oo) == 1
    assert limit(coth(-x), x, oo) == -1


def test_compare():
    assert compare(Integer(2), x, x) < 0
    assert compare(x, exp(x), x) < 0
    assert compare(exp(x), exp(x**2), x) < 0
    assert compare(exp(x**2), exp(exp(x)), x) < 0
    assert compare(Integer(1), exp(exp(x)), x) < 0

    assert compare(x, Integer(2), x) > 0
    assert compare(exp(x), x, x) > 0
    assert compare(exp(x**2), exp(x), x) > 0
    assert compare(exp(exp(x)), exp(x**2), x) > 0
    assert compare(exp(exp(x)), Integer(1), x) > 0

    assert compare(Integer(2), Integer(3), x) == 0
    assert compare(Integer(3), Integer(-5), x) == 0
    assert compare(Integer(2), Integer(-5), x) == 0

    assert compare(x, x**2, x) == 0
    assert compare(x**2, x**3, x) == 0
    assert compare(x**3, 1/x, x) == 0
    assert compare(1/x, x**m, x) == 0
    assert compare(x**m, -x, x) == 0

    assert compare(exp(x), exp(-x), x) == 0
    assert compare(exp(-x), exp(2*x), x) == 0
    assert compare(exp(2*x), exp(x)**2, x) == 0
    assert compare(exp(x)**2, exp(x + exp(-x)), x) == 0
    assert compare(exp(x), exp(x + exp(-x)), x) == 0

    assert compare(exp(x**2), 1/exp(x**2), x) == 0

    assert compare(exp(x), x**5, x) > 0
    assert compare(exp(x**2), exp(x)**2, x) > 0
    assert compare(exp(x), exp(x + exp(-x)), x) == 0
    assert compare(exp(x + exp(-x)), exp(x), x) == 0
    assert compare(exp(x + exp(-x)), exp(-x), x) == 0
    assert compare(exp(-x), x, x) > 0
    assert compare(x, exp(-x), x) < 0
    assert compare(exp(x + 1/x), x, x) > 0
    assert compare(exp(-exp(x)), exp(x), x) > 0
    assert compare(exp(exp(-exp(x)) + x), exp(-exp(x)), x) < 0

    assert compare(exp(exp(x)), exp(x + exp(-exp(x))), x) > 0


def test_signinf():
    assert signinf(Integer(0), x) == 0
    assert signinf(Integer(3), x) == 1
    assert signinf(Integer(-5), x) == -1
    assert signinf(log(x), x) == 1
    assert signinf(exp(-x), x) == 1
    assert signinf(exp(x), x) == 1
    assert signinf(-exp(x), x) == -1
    assert signinf(3 - 1/x, x) == 1
    assert signinf(-3 - 1/x, x) == -1
    assert signinf(sin(1/x), x) == 1
    assert signinf(x**2, x) == 1
    assert signinf(x**5, x) == 1

    assert signinf(x, x) == 1
    assert signinf(-x, x) == -1
    y = Symbol('y', positive=True)
    assert signinf(y, x) == 1
    assert signinf(-y, x) == -1
    assert signinf(y*x, x) == 1
    assert signinf(-y*x, x) == -1


def test_mrv():
    assert mrv(x, x) == {x}
    assert mrv(x + 1/x, x) == {x}
    assert mrv(x**2, x) == {x}
    assert mrv(log(x), x) == {x}
    assert mrv(exp(x), x) == {exp(x)}
    assert mrv(exp(-x), x) == {exp(-x)}
    assert mrv(exp(x**2), x) == {exp(x**2)}
    assert mrv(-exp(1/x), x) == {x}
    assert mrv(exp(x + 1/x), x) == {exp(x + 1/x)}

    assert mrv(exp(x + exp(-exp(x))), x) == {exp(-exp(x))}
    assert mrv(exp(x + exp(-x)), x) == {exp(x + exp(-x)), exp(-x)}
    assert mrv(exp(1/x + exp(-x)), x) == {exp(-x)}

    assert mrv(exp(x + exp(-x**2)), x) == {exp(-x**2)}

    assert mrv(exp(-x + 1/x**2) - exp(x + 1/x), x) == {exp(x + 1/x),
                                                       exp(1/x**2 - x)}

    assert mrv(exp(x**2) + x*exp(x) + exp(x*log(log(x)))/x, x) == {exp(x**2)}
    assert mrv(exp(x)*(exp(1/x + exp(-x)) - exp(1/x)), x) == {exp(x), exp(-x)}
    assert mrv(log(x**2 + 2*exp(exp(3*x**3*log(x)))),
               x) == {exp(exp(3*x**3*log(x)))}
    assert mrv(log(x - log(x))/log(x), x) == {x}
    assert mrv((exp(1/x - exp(-x)) - exp(1/x))*exp(x), x) == {exp(x), exp(-x)}
    assert mrv(1/exp(-x + exp(-x)) - exp(x), x) == {exp(x), exp(-x),
                                                    exp(x - exp(-x))}
    assert mrv(log(log(x*exp(x*exp(x)) + 1)), x) == {exp(x*exp(x))}
    assert mrv(exp(exp(log(log(x) + 1/x))), x) == {x}

    assert mrv((log(log(x) + log(log(x))) -
                log(log(x)))/log(log(x) + log(log(log(x))))*log(x), x) == {x}
    assert mrv(log(log(x*exp(x*exp(x)) + 1)) - exp(exp(log(log(x) + 1/x))),
               x) == {exp(x*exp(x))}

    # Gruntz: p47, 3.21
    h = exp(-x/(1 + exp(-x)))
    e = Mul(exp(h), exp(-x/(1 + h)), exp(exp(-x + h)),
            Pow(h, -2, evaluate=False), evaluate=False) - exp(x) + x
    assert mrv(e, x) == {exp(-x + h), exp(-x/(1 + h)), h, exp(x), exp(-x)}


def test_rewrite():
    assert rewrite(Integer(1), x, m) == (1, None)

    e = exp(x)
    assert rewrite(e, x, m) == (1/m, -x)
    e = exp(x**2)
    assert rewrite(e, x, m) == (1/m, -x**2)
    e = exp(x + 1/x)
    assert rewrite(e, x, m) == (1/m, -x - 1/x)
    e = 1/exp(-x + exp(-x)) - exp(x)
    assert rewrite(e, x, m) == (Add(-1/m, 1/(m*exp(m)), evaluate=False), -x)

    e = exp(x)*log(log(exp(x)))
    assert mrv(e, x) == {exp(x)}
    assert rewrite(e, x, m) == (1/m*log(x), -x)

    e = exp(-x + 1/x**2) - exp(x + 1/x)
    assert rewrite(e, x, m) == (Add(m, Mul(-1, exp(1/x + x**(-2))/m,
                                           evaluate=False),
                                    evaluate=False), -x + 1/x**2)


def test_leadterm():
    assert leadterm(Integer(1), x) == (1, 0)

    assert leadterm(-exp(1/x), x) == (-1, 0)
    assert leadterm(1/exp(-x + exp(-x)) - exp(x), x) == (-1, 0)
    assert leadterm((exp(1/x - exp(-x)) - exp(1/x))*exp(x), x) == (-exp(1/x), 0)

    # Gruntz: p51, 3.25
    assert leadterm((log(exp(x) + x) - x)/log(exp(x) + log(x))*exp(x),
                    x) == (1, 0)

    # Gruntz: p56, 3.27
    assert mrv(exp(-x + exp(-x)*exp(-x*log(x))), x) == {exp(-x*log(x))}
    assert leadterm(exp(-x + exp(-x)*exp(-x*log(x))), x) == (exp(-x), 0)


def test_limit():
    assert limit(x, x, oo) == oo
    assert limit(-x, x, oo) == -oo
    assert limit(-x, x, oo) == -oo
    assert limit((-x)**2, x, oo) == oo
    assert limit(-x**2, x, oo) == -oo
    assert limit((1/x)*log(1/x), x, oo) == 0  # Gruntz: p15, 2.11
    assert limit(1/x, x, oo) == 0  # issue sympy/sympy#11667
    assert limit(exp(x), x, oo) == oo
    assert limit(-exp(x), x, oo) == -oo
    assert limit(exp(x)/x, x, oo) == oo
    assert limit(1/x - exp(-x), x, oo) == 0
    assert limit(x + 1/x, x, oo) == oo
    assert limit(exp(1/log(1 - 1/x)), x, oo) == 0

    assert limit((1/x)**(1/x), x, oo) == 1  # Gruntz: p15, 2.11
    assert limit((exp(1/x) - 1)*x, x, oo) == 1
    assert limit(1 + 1/x, x, oo) == 1
    assert limit(-exp(1/x), x, oo) == -1
    assert limit(x + exp(-x), x, oo) == oo
    assert limit(x + exp(-x**2), x, oo) == oo
    assert limit(x + exp(-exp(x)), x, oo) == oo
    assert limit(13 + 1/x - exp(-x), x, oo) == 13

    assert limit(x - log(1 + exp(x)), x, oo) == 0
    assert limit(x - log(a + exp(x)), x, oo) == 0
    assert limit(exp(x)/(1 + exp(x)), x, oo) == 1
    assert limit(exp(x)/(a + exp(x)), x, oo) == 1

    assert limit((3**x + 5**x)**(1/x), x, oo) == 5  # issue sympy/sympy#3463

    assert limit(Ei(x + exp(-x))*exp(-x)*x, x, oo) == 1

    assert limit(1/li(x), x, oo) == 0
    assert limit(1/Li(x), x, oo) == 0

    # issue diofant/diofant#56
    assert limit((log(E + 1/x) - 1)**(1 - sqrt(E + 1/x)), x, oo) == oo

    # issue sympy/sympy#9471
    assert limit(27**log(x, 3)/x**3, x, oo) == 1
    assert limit(27**(log(x, 3) + 1)/x**3, x, oo) == 27

    # issue sympy/sympy#9449
    assert limit(x*(abs(1/x + y) - abs(y - 1/x))/2, x, oo) == sign(y)**-1

    # issue sympy/sympy#8481
    assert limit(m**x * exp(-m) / factorial(x), x, oo) == 0

    # issue sympy/sympy#4187
    assert limit(exp(1/x)*log(1/x) - Ei(1/x), x, oo) == -EulerGamma
    assert limit(exp(x)*log(x) - Ei(x), x, oo) == oo

    # issue sympy/sympy#10382
    assert limit(fibonacci(x + 1)/fibonacci(x), x, oo) == GoldenRatio

    assert limit(zeta(x), x, oo) == 1
    assert limit(zeta(m)*zeta(x), x, oo) == zeta(m)


def test_I():
    assert limit(I*x, x, oo) == I*oo
    assert limit(y*I*x, x, oo) == sign(y)*I*oo
    assert limit(y*3*I*x, x, oo) == sign(y)*I*oo
    assert limit(y*3*sin(I)*x, x, oo).simplify() == sign(y)*I*oo


def test_sympyissue_4814():
    assert limit((x + 1)**(1/log(x + 1)), x, oo) == E


def test_intractable():
    assert limit(1/gamma(x), x, oo) == 0
    assert limit(1/loggamma(x), x, oo) == 0
    assert limit(gamma(x)/loggamma(x), x, oo) == oo
    assert limit(exp(gamma(x))/gamma(x), x, oo) == oo
    assert limit(gamma(3 + 1/x), x, oo) == 2
    assert limit(gamma(Rational(1, 7) + 1/x), x, oo) == gamma(Rational(1, 7))
    assert limit(log(x**x)/log(gamma(x)), x, oo) == 1
    assert limit(log(gamma(gamma(x)))/exp(x), x, oo) == oo
    assert limit(acosh(1 + 1/x)*sqrt(x), x, oo) == sqrt(2)

    # issue sympy/sympy#10804
    assert limit(2*airyai(x)*root(x, 4) *
                 exp(2*x**Rational(3, 2)/3), x, oo) == 1/sqrt(pi)
    assert limit(airybi(x)*root(x, 4) *
                 exp(-2*x**Rational(3, 2)/3), x, oo) == 1/sqrt(pi)
    assert limit(airyai(1/x), x, oo) == (3**Rational(5, 6) *
                                         gamma(Rational(1, 3))/(6*pi))
    assert limit(airybi(1/x), x, oo) == cbrt(3)*gamma(Rational(1, 3))/(2*pi)
    assert limit(airyai(2 + 1/x), x, oo) == airyai(2)
    assert limit(airybi(2 + 1/x), x, oo) == airybi(2)

    # issue sympy/sympy#10976
    assert limit(erf(m/x)/erf(1/x), x, oo) == m

    assert limit(Max(x**2, x, exp(x))/x, x, oo) == oo


def test_branch_cuts():
    assert limit(sqrt(-1 + I/x), x, oo) == +I
    assert limit(sqrt(-1 - I/x), x, oo) == -I
    assert limit(log(-1 + I/x), x, oo) == +I*pi
    assert limit(log(-1 - I/x), x, oo) == -I*pi
    # Gruntz: p23
    assert limit(atan(2*I - x), x, 0, -1) == I*log(3)/2 - pi/2
    assert limit(atan(2*I - x), x, 0, +1) == I*log(3)/2 + pi/2


def test_aseries_trig():
    assert limit(1/log(atan(x)), x, oo) == -1/(-log(pi) + log(2))
    assert limit(1/acot(-x), x, oo) == -oo


def test_exp_log_series():
    assert limit(x/log(log(x*exp(x))), x, oo) == oo


def test_sympyissue_3644():
    assert limit(((x**7 + x + 1)/(2**x + x**2))**(-1/x), x, oo) == 2


def test_sympyissue_6843():
    n = Symbol('n', integer=True, positive=True)
    r = (n + 1)*(1 + 1/x)**(n + 1)/((1 + 1/x)**(n + 1) - 1) - (1 + 1/x)*x
    assert limit(r, x, oo) == n/2


def test_sympyissue_4190():
    assert limit(x - gamma(1/x), x, oo) == EulerGamma


def test_sympyissue_4109():
    assert limit(1/gamma(1/x), x, oo) == 0
    assert limit(gamma(1/x)/x, x, oo) == 1


def test_sympyissue_6682():
    assert limit(exp(2*Ei(-1/x))*x**2, x, oo) == exp(2*EulerGamma)


def test_sympyissue_7096():
    assert limit((-1/x)**-pi, x, oo) == oo*sign((-1)**(-pi))


def test_sympyissue_8462():
    assert limit(binomial(x, x/2), x, oo) == oo
    # issue sympy/sympy#10801
    assert limit(16**x/(x*binomial(2*x, x)**2), x, oo) == pi


def test_issue_74():
    assert limit(sign(log(1 + 1/x)), x, oo) == +1
    assert limit(sign(log(1 - 1/x)), x, oo) == -1
    assert limit(sign(sin(+1/x)), x, oo) == +1
    assert limit(sign(sin(-1/x)), x, oo) == -1
    assert limit(sign(tan(+1/x)), x, oo) == +1
    assert limit(sign(tan(-1/x)), x, oo) == -1
    assert limit(sign(cos(pi/2 + 1/x)), x, oo) == -1
    assert limit(sign(cos(pi/2 - 1/x)), x, oo) == +1


def test_issue_75():
    assert limit(abs(log(x)), x, oo) == oo
    assert limit(tan(abs(pi/2 + 1/x))/acosh(pi/2 + 1/x), x, oo) == -oo
    assert limit(tan(abs(pi/2 - 1/x))/acosh(pi/2 - 1/x), x, oo) == +oo

    assert limit(abs(log(2 + 1/x)) - log(2 + 1/x), x, oo) == 0
    assert limit(abs(log(2 - 1/x)) - log(2 - 1/x), x, oo) == 0


def test_sympyissue_8241():
    e = x/log(x)**(log(x)/(m*log(log(x))))
    assert isinstance(limit(e, x, oo), Limit)


def test_sympyissue_23845():
    e = ((-sqrt(5)*(-sqrt(5)/2 + Rational(3, 2))**x +
          5*(-sqrt(5)/2 + Rational(3, 2))**x +
          (sqrt(5)/2 + Rational(3, 2))**x*(sqrt(5) + 5)) /
         (-3*sqrt(5)*(-sqrt(5)/2 + Rational(3, 2))**x +
          5*(-sqrt(5)/2 + Rational(3, 2))**x +
          (5 + 3*sqrt(5))*(sqrt(5)/2 + Rational(3, 2))**x))
    assert limit(e, x, oo) == (sqrt(5) + 5)/(5 + 3*sqrt(5))



"""Efficient functions for generating orthogonal polynomials."""

from ..core import Dummy
from ..domains import QQ, ZZ
from .constructor import construct_domain
from .polytools import Poly, PurePoly


def _jacobi(n, a, b, K):
    """Low-level implementation of Jacobi polynomials."""
    ring = K.poly_ring('_0')
    x = ring._0

    j0 = ring.one
    if n < 1:
        return j0
    j1 = ((a + b + K(2))*x + (a - b))/K(2)

    for i in range(2, n + 1):
        den = K(i)*(a + b + i)*(a + b + K(2)*i - K(2))
        f0 = (a + b + K(2)*i - K.one) * (a*a - b*b) / (K(2)*den)
        f1 = (a + b + K(2)*i - K.one) * (a + b + K(2)*i - K(2)) * (a + b + K(2)*i) / (K(2)*den)
        f2 = (a + i - K.one)*(b + i - K.one)*(a + b + K(2)*i) / den
        j0, j1 = j1, j1*f0 + j1*x*f1 - j0*f2

    return j1


def jacobi_poly(n, a, b, x=None, **args):
    """Generates Jacobi polynomial of degree `n` in `x`."""
    if n < 0:
        raise ValueError(f"can't generate Jacobi polynomial of degree {n}")

    K, v = construct_domain([a, b], field=True)
    poly = _jacobi(int(n), v[0], v[1], K)

    if x is not None:
        poly = Poly(poly, x, domain=K)
    else:
        poly = PurePoly(poly, Dummy('x'), domain=K)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly


def _gegenbauer(n, a, K):
    """Low-level implementation of Gegenbauer polynomials."""
    ring = K.poly_ring('_0')
    x = ring._0

    g0 = ring.one
    if n < 1:
        return g0
    g1 = K(2)*a*x

    for i in range(2, n + 1):
        f1 = K(2) * (i + a - K.one) / i
        f2 = (i + K(2)*a - K(2)) / i
        g0, g1 = g1, g1*x*f1 - g0*f2

    return g1


def gegenbauer_poly(n, a, x=None, **args):
    """Generates Gegenbauer polynomial of degree `n` in `x`."""
    if n < 0:
        raise ValueError(
            f"can't generate Gegenbauer polynomial of degree {n}")

    K, a = construct_domain(a, field=True)
    poly = _gegenbauer(int(n), a, K)

    if x is not None:
        poly = Poly(poly, x, domain=K)
    else:
        poly = PurePoly(poly, Dummy('x'), domain=K)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly


def _chebyshevt(n, K):
    """Low-level implementation of Chebyshev polynomials of the 1st kind."""
    ring = K.poly_ring('_0')
    x = ring._0

    c0 = ring.one
    if n < 1:
        return c0
    c1 = x

    for _ in range(2, n + 1):
        a = c1*x*K(2)
        c0, c1 = c1, a - c0

    return c1


def chebyshevt_poly(n, x=None, **args):
    """Generates Chebyshev polynomial of the first kind of degree `n` in `x`."""
    if n < 0:
        raise ValueError(
            f"can't generate 1st kind Chebyshev polynomial of degree {n}")

    poly = _chebyshevt(int(n), ZZ)

    if x is not None:
        poly = Poly(poly, x, domain=ZZ)
    else:
        poly = PurePoly(poly, Dummy('x'), domain=ZZ)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly


def _chebyshevu(n, K):
    """Low-level implementation of Chebyshev polynomials of the 2nd kind."""
    ring = K.poly_ring('_0')
    x = ring._0

    seq = [ring.one, K(2)*x]

    for _ in range(2, n + 1):
        a = seq[-1]*x*K(2)
        seq.append(a - seq[-2])

    return seq[n]


def chebyshevu_poly(n, x=None, **args):
    """Generates Chebyshev polynomial of the second kind of degree `n` in `x`."""
    if n < 0:
        raise ValueError(
            f"can't generate 2nd kind Chebyshev polynomial of degree {n}")

    poly = _chebyshevu(int(n), ZZ)

    if x is not None:
        poly = Poly(poly, x, domain=ZZ)
    else:
        poly = PurePoly(poly, Dummy('x'), domain=ZZ)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly


def _hermite(n, K):
    """Low-level implementation of Hermite polynomials."""
    ring = K.poly_ring('_0')
    x = ring._0

    h0 = ring.one
    if n < 1:
        return h0
    h1 = K(2)*x

    for i in range(2, n + 1):
        a = h1*x
        b = h0*K(i - 1)

        h0, h1 = h1, (a - b)*K(2)

    return h1


def hermite_poly(n, x=None, **args):
    """Generates Hermite polynomial of degree `n` in `x`."""
    if n < 0:
        raise ValueError(f"can't generate Hermite polynomial of degree {n}")

    poly = _hermite(int(n), ZZ)

    if x is not None:
        poly = Poly(poly, x, domain=ZZ)
    else:
        poly = PurePoly(poly, Dummy('x'), domain=ZZ)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly


def _legendre(n, K):
    """Low-level implementation of Legendre polynomials."""
    ring = K.poly_ring('_0')
    x = ring._0

    l0 = ring.one
    if n < 1:
        return l0
    l1 = x

    for i in range(2, n + 1):
        l0, l1 = l1, l1*x*K(2*i - 1, i) - l0*K(i - 1, i)

    return l1


def legendre_poly(n, x=None, **args):
    """Generates Legendre polynomial of degree `n` in `x`."""
    if n < 0:
        raise ValueError(f"can't generate Legendre polynomial of degree {n}")

    poly = _legendre(int(n), QQ)

    if x is not None:
        poly = Poly(poly, x, domain=QQ)
    else:
        poly = PurePoly(poly, Dummy('x'), domain=QQ)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly


def _laguerre(n, alpha, K):
    """Low-level implementation of Laguerre polynomials."""
    ring = K.poly_ring('_0')
    x = ring._0

    l0, l1 = ring.zero, ring.one

    for i in range(1, n + 1):
        l0, l1 = l1, l1*(-K.one/i*x + alpha/i + K(2*i - 1)/i) - l0*(alpha/i + K(i - 1)/i)

    return l1


def laguerre_poly(n, x=None, alpha=None, **args):
    """Generates Laguerre polynomial of degree `n` in `x`."""
    if n < 0:
        raise ValueError(f"can't generate Laguerre polynomial of degree {n}")

    if alpha is not None:
        K, alpha = construct_domain(
            alpha, field=True)  # XXX: ground_field=True
    else:
        K, alpha = QQ, QQ(0)

    poly = _laguerre(int(n), alpha, K)

    if x is not None:
        poly = Poly(poly, x, domain=K)
    else:
        poly = PurePoly(poly, Dummy('x'), domain=K)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly


def _spherical_bessel_fn(n, K):
    """Low-level implementation of fn(n, x)."""
    ring = K.poly_ring('_0')
    x = ring._0

    s0 = ring.one
    if n < 1:
        return s0*x
    s1 = x

    for i in range(2, n + 1):
        s0, s1 = s1, s1*x*K(2*i - 1) - s0

    return s1*x


def _spherical_bessel_fn_minus(n, K):
    """Low-level implementation of fn(-n, x)."""
    ring = K.poly_ring('_0')
    x = ring._0

    s0, s1 = x, ring.zero

    for i in range(2, n + 1):
        s0, s1 = s1, s1*x*K(3 - 2*i) - s0

    return s1


def spherical_bessel_fn(n, x=None, **args):
    """
    Coefficients for the spherical Bessel functions.

    Those are only needed in the jn() function.

    The coefficients are calculated from:

    fn(0, z) = 1/z
    fn(1, z) = 1/z**2
    fn(n-1, z) + fn(n+1, z) == (2*n+1)/z * fn(n, z)

    Examples
    ========

    >>> spherical_bessel_fn(1, z)
    z**(-2)
    >>> spherical_bessel_fn(2, z)
    -1/z + 3/z**3
    >>> spherical_bessel_fn(3, z)
    -6/z**2 + 15/z**4
    >>> spherical_bessel_fn(4, z)
    1/z - 45/z**3 + 105/z**5

    """
    if n < 0:
        poly = _spherical_bessel_fn_minus(-int(n), ZZ)
    else:
        poly = _spherical_bessel_fn(int(n), ZZ)

    if x is not None:
        poly = Poly(poly, 1/x, domain=ZZ)
    else:
        poly = PurePoly(poly, 1/Dummy('x'), domain=ZZ)

    if not args.get('polys', False):
        return poly.as_expr()
    else:
        return poly
